/*
   simple tool for a parallel search of a simple string within a file
*/


#include <cassert>
#include <cstdlib>
#include <iostream>
#include <mutex>
#include <thread>
#include <set>
#include <vector>
#include <string>
#include <sstream>
#include <string.h>
#include "mapped_file.hpp"

struct Result {
   Result(unsigned int tid_, unsigned int rline_,
	 const char* line_, size_t len_) :
      tid(tid_), rline(rline_), line(line_), len(len_) {
   }
   unsigned int tid; // thread id
   unsigned int rline; // relative line number
   const char* line; // pointer to the begin of the corresponding line
   size_t len; // length of the line
};

bool operator<(const Result& r1, const Result& r2) {
   return r1.line < r2.line; // just compare the pointers
}

class Results {
   public:
      Results(unsigned int nthreads) : nofthreads(nthreads), lines(nofthreads) {
      }
      void add(const Result& result) {
	 std::lock_guard<std::mutex> lock(mutex);
	 results.insert(result);
      }
      void set_lines(unsigned int tid, unsigned int noflines) {
	 std::lock_guard<std::mutex> lock(mutex);
	 assert(tid < nofthreads);
	 lines[tid] = noflines;
      }
      void print() const {
	 std::lock_guard<std::mutex> lock(mutex);
	 std::vector<unsigned int> offsets(lines);
	 unsigned last_offset = 1; /* lines are numbered from 1 */
	 for (auto& offset: offsets) {
	    auto count = offset; offset = last_offset; last_offset += count;
	 }
	 for (auto& result: results) {
	    unsigned int line = offsets[result.tid] + result.rline;
	    std::cout << line << ": ";
	    std::cout.write(result.line, result.len);
	    std::cout << std::endl;
	 }
      }
   private:
      unsigned int nofthreads;
      mutable std::mutex mutex;
      std::set<Result> results;
      std::vector<unsigned int> lines;
};

int main(int argc, char** argv) {
   const char* cmdname = *argv++; --argc;
   if (argc != 3) {
      std::cerr << "Usage: " << cmdname << " nthreads string file" << std::endl;
      exit(1);
   }
   const char* nthreads = *argv++; --argc;
   const char* string = *argv++; --argc;
   size_t string_len = strlen(string);
   if (string_len == 0) {
      std::cerr << cmdname << ": empty search string given" << std::endl;
      exit(1);
   }
   const char* filename = *argv++; --argc;
   
   std::string nthreads_s(nthreads);
   std::istringstream is(nthreads_s);
   unsigned int nofthreads;
   if (!(is >> nofthreads) || nofthreads == 0) {
      std::cerr << cmdname << ": bad value for nthreads" << std::endl;
      exit(1);
   }

   MappedFile mf(filename);
   if (!mf.valid()) {
      std::cerr << cmdname << ": " << filename << ": " <<
	 strerror(mf.get_errno()) << std::endl;
      exit(1);
   }
   const char* addr = mf.get_addr();
   size_t len = mf.get_length();

   if (nofthreads >= len) {
      nofthreads = 1;
   }

   const char* tstart[nofthreads];
   size_t tlen[nofthreads];
   const char* start = addr;
   for (unsigned int i = 0; i < nofthreads; ++i) {
      tstart[i] = start;
      if (i == nofthreads - 1) {
	 tlen[i] = len - (start - addr); /* take the rest */
      } else {
	 const char* cp = start + (addr + len - start) / (nofthreads - i);
	 /* find next line terminator */
	 while (cp < addr + len && *cp != '\n') ++cp;
	 if (cp < addr) ++cp; /* skip line terminator */
	 tlen[i] = cp - start;
	 start = cp;
	 if (cp == addr + len) {
	    /* we need less threads */
	    nofthreads = i + 1; break;
	 }
      }
   }
   Results results(nofthreads);
   std::thread t[nofthreads];
   for (unsigned int tid = 0; tid < nofthreads; ++tid) {
      const char* addr = tstart[tid]; size_t len = tlen[tid];
      t[tid] = std::thread([=,&results]() {
	 bool matches[string_len]; memset(matches, 0, sizeof(matches));
	 unsigned int index = 0;
	 unsigned int lines = 0; const char* current_line = 0;
	 bool found = false; int partial = 0;
	 for (const char* cp = addr; cp < addr + len; ++cp) {
	    if (!current_line) {
	       current_line = cp;
	    }
	    if (*cp == '\n') {
	       if (found) {
		  results.add(Result(tid, lines, current_line,
		     cp - current_line));
		  found = false;
	       }
	       ++lines;
	       /* reset engine */
	       current_line = 0; partial = 0;
	       index = 0; memset(matches, 0, sizeof(matches));
	       continue;
	    }
	    if (found) continue;
	    if (!partial && *cp != string[0]) continue;
	    for (unsigned int i = 0; i < string_len; ++i) {
	       unsigned int j = (index + string_len - i) % string_len;
	       if (*cp == string[i]) {
		  if (i == 0) {
		     matches[j] = true; ++partial;
		  }
		  if (i == string_len - 1 && matches[j]) {
		     found = true;
		  }
	       } else if (matches[j]) {
		  matches[j] = false; --partial;
	       }
	    }
	    index = (index + 1) % string_len;
	 }
	 results.set_lines(tid, lines);
      });
   }
   for (auto& thr: t) {
      thr.join();
   }
   results.print();
}
